import pytest
import os
import xarray as xr
import numpy as np
import pickle
from yadg.extractors.eclab.mpr import extract as extract_mpr
from yadg.extractors.eclab.mpt import extract as extract_mpt
from .utils import compare_datatrees
from pathlib import Path


def check_file(fname, kwargs, func):
    ret = func(source=Path(fname), **kwargs)
    outfile = f"{fname}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)
    return ret


def check_file_extract_bytes(fname, bytes, kwargs, func):
    ret = func(source=bytes, **kwargs)
    outfile = f"{fname}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)
    return ret


def compare_params(left, right):
    lpar = left.attrs["original_metadata"].get("params", {})
    rpar = right.attrs["original_metadata"].get("params", {})
    for k in rpar:
        if k.endswith("vs.") or k.startswith("unit") or k == "I Range":
            assert k in lpar, f"Param {k!r} not found in mpt file: {lpar.keys()}"
            assert lpar[k] == rpar[k], f"Inconsistent param {k!r}: {lpar[k]}, {rpar[k]}"


@pytest.mark.parametrize(
    "froot, locale",
    [
        ("bcd.issue_241", "en_US"),
        ("ca", "en_US"),
        ("ca.issue_134", "en_US"),
        ("ca.issue_149", "de_DE"),
        ("coc.issue_185", "en_US"),
        ("cov.issue_185", "en_US"),
        ("cp", "en_US"),
        ("cp.issue_61", "en_US"),
        ("cp.issue_149", "de_DE"),
        ("cv", "en_US"),
        ("cv.issue_149", "de_DE"),
        ("cv.issue_217", "de_DE"),
        ("cva.issue_135", "en_US"),
        ("cva.issue_202", "en_US"),
        ("gcpl", "en_US"),
        ("gcpl.issue_149", "de_DE"),
        ("gcpl.issue_175", "en_US"),
        ("gcpl.pr_182.1", "en_US"),
        ("gcpl.pr_182.2", "en_US"),
        ("gcpl.issue_211", "en_US"),
        ("gcpl.issue_226.I", "en_US"),
        ("gcpl.issue_226.CxN", "en_US"),
        ("gcpl.issue_228", "en_US"),
        ("gcpl.issue_230", "en_US"),
        ("geis", "en_US"),
        ("lsv", "en_US"),
        ("lsv.issue_195", "en_US"),
        ("mb", "en_US"),
        ("mb.issue_95", "en_US"),
        ("mb.issue_149", "de_DE"),
        ("mb.issue_218", "en_US"),
        ("mb.issue_223", "en_GB"),
        ("mp.issue_183", "en_US"),
        ("ocv", "en_US"),
        ("ocv.issue_149", "de_DE"),
        ("peis", "en_US"),
        ("peis.issue_225.no_ece", "en_US"),
        ("peis.issue_225.with_ece", "en_US"),
        ("peis.issue_225.ewe_ece", "en_US"),
        ("wait", "en_US"),
        ("vsp_ocv_wo", "en_US"),
        ("vsp_ocv_with", "en_US"),
        ("vsp_peis_with", "en_US"),
        ("zir", "en_US"),
    ],
)
def test_eclab_consistency(froot, locale, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", locale=locale, encoding="windows-1252")
    aret = check_file(f"{froot}.mpr", kwargs, extract_mpr)
    bret = check_file(f"{froot}.mpt", kwargs, extract_mpt)

    for key in aret.variables:
        if key.endswith("std_err"):
            continue

        bkey = key
        if bkey not in bret:
            if key.replace("we", "") in bret:
                bkey = key.replace("we", "")
            elif key.replace("we ", "") in bret:
                bkey = key.replace("we ", "")
        assert bkey in bret

        try:
            if np.isdtype(aret[key].dtype, "numeric"):
                np.testing.assert_allclose(
                    aret[key],
                    bret[bkey],
                    equal_nan=True,
                    atol=1e-8,
                    rtol=1e-5,
                )
            else:
                np.testing.assert_array_equal(aret[key], bret[bkey])
        except AssertionError as e:
            e.args = (e.args[0] + f"\nError happened on key: {key!r}\n",)
            raise e
    compare_params(aret, bret)


@pytest.mark.parametrize(
    "froot, locale",
    [
        ("ca", "en_US"),
    ],
)
def test_eclab_consistency_extract_mpr_bytes(froot, locale, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", locale=locale, encoding="windows-1252")
    with open(f"{froot}.mpr", "rb") as mpr_file:
        mpr_bytes = mpr_file.read()
    aret = check_file_extract_bytes(f"{froot}.mpr", mpr_bytes, kwargs, extract_mpr)
    bret = check_file(f"{froot}.mpr", kwargs, extract_mpr)

    for key in aret.variables:
        try:
            xr.testing.assert_equal(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"\nError happened on key: {key!r}\n",)
            raise e
    compare_params(aret, bret)


@pytest.mark.parametrize(
    "froot, locale",
    [
        ("geis.issue_149", "en_US"),
        ("peis.issue_149", "en_US"),
    ],
)
def test_eclab_consistency_partial_149(froot, locale, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", locale=locale, encoding="windows-1252")
    aret = check_file(f"{froot}.mpr", kwargs, extract_mpr)
    bret = check_file(f"{froot}.mpt", kwargs, extract_mpt)

    for key in aret.variables:
        if key.endswith("std_err"):
            continue
        elif key in {
            "THD Ewe",
            "NSD Ewe",
            "NSR Ewe",
            "|Ewe h2|",
            "|Ewe h3|",
            "|Ewe h4|",
            "|Ewe h5|",
            "|Ewe h6|",
            "|Ewe h7|",
            "THD I",
            "NSD I",
            "NSR I",
            "|I h2|",
            "|I h3|",
            "|I h4|",
            "|I h5|",
            "|I h6|",
            "|I h7|",
        }:
            continue

        bkey = key
        if bkey not in bret:
            if key.replace("we", "") in bret:
                bkey = key.replace("we", "")
            elif key.replace("we ", "") in bret:
                bkey = key.replace("we ", "")

        assert bkey in bret

        try:
            if np.isdtype(aret[key].dtype, "numeric"):
                np.testing.assert_allclose(
                    aret[key],
                    bret[bkey],
                    equal_nan=True,
                    atol=1e-8,
                    rtol=1e-5,
                )
            else:
                np.testing.assert_array_equal(aret[key], bret[bkey])
        except AssertionError as e:
            e.args = (e.args[0] + f"\nError happened on key: {key!r}\n",)
            raise e
    compare_params(aret, bret)


@pytest.mark.parametrize(
    "afile, bfile",
    [
        ("mb.issue_95.mpt", "mb.issue_95.de.mpt"),
    ],
)
def test_eclab_mpt_locale(afile, bfile, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", encoding="windows-1252")
    aret = extract_mpt(source=Path(afile), locale="en_US", **kwargs)
    bret = extract_mpt(source=Path(bfile), locale="de_DE", **kwargs)
    compare_datatrees(aret, bret)


@pytest.mark.parametrize(
    "afile, bfile",
    [
        ("mb.old.mpt", "mb.mpt"),
        ("gcpl.old.mpt", "gcpl.mpt"),
        ("ocv.old.mpt", "ocv.mpt"),
        ("wait.old.mpt", "wait.mpt"),
    ],
)
def test_eclab_mpt_old(afile, bfile, datadir):
    os.chdir(datadir)
    kwargs = dict(timezone="Europe/Berlin", encoding="windows-1252")
    aret = extract_mpt(source=Path(afile), locale="en_US", **kwargs)
    bret = extract_mpt(source=Path(bfile), locale="en_US", **kwargs)

    for key in aret.variables:
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            if key in {"Efficiency", "Efficiency_std_err"}:
                pass
            else:
                e.args = (e.args[0] + f"\nError happened on key: {key!r}\n",)
                raise e



import pytest
import os
import json
import yadg.dgutils
import yadg.core
import xarray as xr
from .utils import compare_datatrees


@pytest.mark.parametrize(
    "input",
    [
        "test_locale_passthrough.json",
    ],
)
def test_process_locale_passthrough(input, datadir):
    os.chdir(datadir)

    with open(f"{input}") as infile:
        obj = json.load(infile)

    schema = yadg.dgutils.update_schema(obj)
    ret = yadg.core.process_schema(schema)
    # ret.to_netcdf(f"{input}.nc", engine="h5netcdf")
    ref = xr.open_datatree(f"{input}.nc", engine="h5netcdf")
    compare_datatrees(ret, ref, thislevel=True, descend=True)



import pytest
import os
import pickle
from yadg.extractors.panalytical.csv import extract
from .utils import compare_datatrees
from pathlib import Path


@pytest.mark.parametrize(
    "infile",
    [
        "210520step1_30min.csv",
    ],
)
def test_panalytical_csv(infile, datadir):
    os.chdir(datadir)
    ret = extract(source=Path(infile), encoding="utf-8", timezone="Europe/Berlin")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



import pytest
import os
import shutil
from yadg.extractors.panalytical.xrdml import extract as extract_xrdml
from yadg.extractors.panalytical.csv import extract as extract_csv
import xarray as xr
from pathlib import Path


@pytest.fixture
def _datadir(tmpdir, request):
    filename = request.module.__file__
    test_dir, _ = os.path.splitext(filename)
    base_dir, _ = os.path.split(test_dir)
    for ddir in ["test_x_panalytical_xrdml", "test_x_panalytical_csv"]:
        shutil.copytree(os.path.join(base_dir, ddir), str(tmpdir), dirs_exist_ok=True)
    return tmpdir


@pytest.mark.parametrize(
    "afile, bfile",
    [
        ("210520step1_30min.csv", "210520step1_30min.xrdml"),
    ],
)
def test_panalytical_consistency(afile, bfile, _datadir):
    os.chdir(_datadir)
    aret = extract_csv(source=Path(afile), encoding="utf-8", timezone="Europe/Berlin")
    bret = extract_xrdml(source=Path(bfile), encoding="utf-8", timezone="Europe/Berlin")
    for key in aret.variables:
        if key.endswith("std_err"):
            continue
        try:
            xr.testing.assert_allclose(aret[key], bret[key])
        except AssertionError as e:
            e.args = (e.args[0] + f"Error happened on key: {key!r}\n",)
            raise e



import pytest
import os
import pickle
from yadg.extractors.fhimcpt.csv import extract
from dgbowl_schemas.yadg.dataschema_6_0.filetype import FHI_csv
from .utils import compare_datatrees
from pathlib import Path


@pytest.mark.parametrize(
    "infile",
    [
        "measurement.csv",
    ],
)
def test_fhimcpt_csv(infile, datadir):
    os.chdir(datadir)
    ret = extract(
        source=Path(infile),
        parameters=FHI_csv(filetype="fhimcpt.csv", parameters={}).parameters,
        encoding="utf8",
        locale="en_GB",
        timezone="Europe/Berlin",
    )
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



import importlib
import logging
import json
from functools import wraps
from functools import singledispatch
from typing import Any
from pathlib import Path
from xarray import DataTree
from yadg import dgutils
from dgbowl_schemas.yadg.dataschema import ExtractorFactory, FileType


logger = logging.getLogger(__name__)


def deprecate_fn_path(func):
    def handle_deprecated(param, *args, **kwargs):
        logger.warning(
            f"The parameter '{param}' is deprecated and has been replaced by 'source'. ({DeprecationWarning.__name__})",
        )
        source = kwargs.get("source", kwargs.pop(param))
        args = (source,) + args
        return args, kwargs

    @wraps(func)
    def wrapper(*args, **kwargs):
        if kwargs.get("source") is not None:
            args = (kwargs.pop("source"),) + args
        elif kwargs.get("fn") is not None:
            args, kwargs = handle_deprecated("fn", *args, **kwargs)
        elif kwargs.get("path") is not None:
            args, kwargs = handle_deprecated("path", *args, **kwargs)
        return func(*args, **kwargs)

    return wrapper


def extract(
    filetype: str,
    path: Path | str,
    timezone: str = None,
    encoding: str = None,
    locale: str = None,
    **kwargs: dict,
) -> DataTree:
    """
    Extract data and metadata from a path using the supplied filetype.

    A wrapper around the :func:`extract_from_path` worker function, which creates a
    default extractor object. Coerces any :class:`str`s provided as ``path`` to :class:`Path`.

    Parameters
    ----------

    filetype:
        Specifies the filetype. Has to be a filetype supported by the dataschema.

    path:
        A :class:`Path` object pointing to the file to be extracted.

    timezone:
        A :class:`str` containing the TZ identifier, e.g. "Europe/Berlin".

    encoding:
        A :class:`str` containing the encoding, e.g. "utf-8".

    locale:
        A :class:`str` containing the locale name, e.g. "de_CH".

    """
    extractor = ExtractorFactory(
        extractor={
            "filetype": filetype,
            "timezone": timezone,
            "encoding": encoding,
            "locale": locale,
        }
    ).extractor

    return extract_from_path(Path(path), extractor, **kwargs)


@deprecate_fn_path
def extract_from_path(
    source: Path,
    extractor: FileType,
    **kwargs: dict,
) -> DataTree:
    """
    Extracts data and metadata from the provided path using the supplied extractor.

    The individual extractor functionality of yadg is called from here. The data is
    always returned as a :class:`DataTree`. The ``original_metadata`` entries in
    the returned objects are flattened using json serialisation. The returned objects
    have a :func:`to_netcdf` as well as a :func:`to_dict` method, which can be used to
    write the returned object into a file.

    Parameters
    ----------

    source:
        A :obj:`Path` pointing to the extracted file.

    extractor:
        A :class:`FileType` object describing the extraction process.
    """

    m = importlib.import_module(f"yadg.extractors.{extractor.filetype}")
    func = getattr(m, "extract")

    # Func should always return a xarray.DataTree
    ret: DataTree = func(source=source, **vars(extractor), **kwargs)
    jsonize_orig_meta(ret)

    ret.attrs.update(
        {
            "yadg_provenance": "yadg extract",
            "yadg_extract_date": dgutils.now(asstr=True),
            "yadg_extract_filename": str(source),
            "yadg_extract_Extractor": extractor.model_dump_json(exclude_none=True),
        }
    )
    ret.attrs.update(dgutils.get_yadg_metadata())

    return ret


def extract_from_bytes(
    source: bytes,
    extractor: FileType,
    **kwargs: dict,
) -> DataTree:
    """
    Extracts data and metadata from the provided raw bytes using the supplied extractor.

    The individual extractor functionality of yadg is called from here. The data is
    always returned as a :class:`DataTree`. The ``original_metadata`` entries in
    the returned objects are flattened using json serialisation. The returned objects
    have a :func:`to_netcdf` as well as a :func:`to_dict` method, which can be used to
    write the returned object into a file.

    Parameters
    ----------

    source:
        A :obj:`bytes` object containing the raw data to be extracted.

    extractor:
        A :class:`FileType` object describing the extraction process.
    """

    m = importlib.import_module(f"yadg.extractors.{extractor.filetype}")
    func = getattr(m, "extract")

    # Func should always return a xarray.DataTree
    ret: DataTree = func(source=source, **vars(extractor))
    jsonize_orig_meta(ret)

    ret.attrs.update(
        {
            "yadg_provenance": "yadg extract",
            "yadg_extract_date": dgutils.now(asstr=True),
            "yadg_extract_Extractor": extractor.model_dump_json(exclude_none=True),
        }
    )
    ret.attrs.update(dgutils.get_yadg_metadata())

    return ret


def jsonize_orig_meta(obj: DataTree):
    for k in obj:
        if isinstance(obj[k], DataTree):
            jsonize_orig_meta(obj[k])
    if "original_metadata" in obj.attrs:
        obj.attrs["original_metadata"] = json.dumps(obj.attrs["original_metadata"])
    if "fulldate" in obj.attrs:
        obj.attrs["fulldate"] = int(obj.attrs["fulldate"])


__all__ = ["extract"]


def get_extract_dispatch():
    @deprecate_fn_path
    @singledispatch
    def extract(
        source: Any,
        *,
        timezone: str,
        **kwargs: dict,
    ) -> DataTree:
        raise RuntimeError(
            f"The selected extractor does not support the source type {type(source)!r} "
            "provided. Please check the available extractors or enter a valid file path."
        )

    return extract



import pytest
import os
import pickle
from yadg.extractors.fusion.json import extract
from .utils import compare_datatrees
from pathlib import Path


@pytest.mark.parametrize(
    "infile",
    [
        "15p-Cu-10mA-01 - Jun 08 2022, 16;10.fusion-data",
        "15p-Cu-10mA-01 - Jun 08 2022, 16;23.fusion-data",
        "AgPTFE28_100mA_NaS_01 - Aug 13 2021, 17;56.fusion-data",
        "AgPTFE28_100mA_NaS_01 - Aug 13 2021, 18;18.fusion-data",
        "issue_198.fusion-data",
        "issue_213.fusion-data",
        "issue_249.fusion-data",
    ],
)
def test_fusion_json(infile, datadir):
    os.chdir(datadir)
    ret = extract(source=Path(infile), encoding="utf-8", timezone="Europe/Berlin")
    outfile = f"{infile}.pkl"
    with open(outfile, "rb") as inp:
        ref = pickle.load(inp)
    print(f"{ret=}")
    with open(outfile, "wb") as out:
        pickle.dump(ret, out, 5)
    compare_datatrees(ret, ref, thislevel=True)



"""
Handles processing of Panalytical XRDML files.

Usage
`````
Available since ``yadg-4.2``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_6_0.filetype.Panalytical_xrdml

Schema
``````
.. code-block:: yaml

    xarray.DataTree:
      coords:
        uts:            !!float               # Unix timestamp
        angle:          !!float               # 2θ angle
      data_vars:
        intensity:      (uts, angle)          # Measured intensity

Metadata
````````
The following metadata is extracted:

  - ``sample``: Metadata information about the sample.
  - ``wavelength``: Measurement wavelength.
  - ``comment``: A free-form description of the experiment.
  - ``incident_beam_path``
  - ``diffracted_beam_path``
  - ``counting_time``

.. note::

    The returned metadata contain only a subset of the available metadata in the XML
    file. If something important is missing, please contact us!

Notes on file structure
```````````````````````
These are xml-formatted files, which we here parse using the :mod:`xml.etree`
library into a Python :class:`dict`.

The ``angle`` returned from this parser is based on a linear interpolation of the start
and end point of the scan, and is the :math:`2\\theta`. The values of :math:`\\omega`
are discarded.

Uncertainties
`````````````
The uncertainties of in ``"angle"`` are taken as the step-width of the linearly spaced
:math:`2\\theta` values.

The uncertainties of of ``"intensity"`` are currently set to a constant
value of 1.0 count as all the supported files seem to produce integer values.

.. codeauthor::
    Nicolas Vetsch,
    Peter Kraus

"""

from collections import defaultdict
from xml.etree import ElementTree
import numpy as np
from xarray import DataTree
import xarray as xr
from uncertainties.core import str_to_number_with_uncert as tuple_fromstr

from yadg.extractors.panalytical.common import panalytical_comment
from yadg.dgutils import dateutils
from pathlib import Path
from yadg.extractors import get_extract_dispatch

extract = get_extract_dispatch()


def etree_to_dict(e: ElementTree.Element) -> dict:
    """Recursively converts an ElementTree.Element into a dictionary.

    Element attributes are stored into `"@"`-prefixed attribute keys.
    Element text is stored into `"#text"` for all nodes.

    From https://stackoverflow.com/a/10076823.
    """
    d = {e.tag: {} if e.attrib else None}
    children = list(e)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d = {e.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
    if e.attrib:
        d[e.tag].update(("@" + k, v) for k, v in e.attrib.items())
    if e.text:
        text = e.text.strip()
        if children or e.attrib:
            if text:
                d[e.tag]["#text"] = text
        else:
            d[e.tag] = text
    return d


def _process_values(d: dict | str) -> dict | str:
    """
    Recursively parses dicts in the following format:

    .. code::

        {"key": {"#text": ..., "@unit": ...}, ...}

    into a single string:

    .. code::

        {"key": f"{#text} {@unit}", ...}
    """
    # TODO
    # If not "#text" or @tribute just snake_case and recurse.
    if isinstance(d, dict):
        if "@unit" in d and "#text" in d:
            return f"{d['#text']} {d['@unit']}"
        elif "@version" in d and "#text" in d:
            return f"{d['#text']} {d['@version']}"
        else:
            for k, v in d.items():
                d[k] = _process_values(v)
    return d


def _process_scan(scan: dict) -> dict:
    """
    Parses the scan section of the file. Creates the explicit positions based
    on the number of measured intensities and the start & end position.
    """
    header = scan.pop("header")
    dpts = scan.pop("dataPoints")
    counting_time = _process_values(dpts.pop("commonCountingTime"))
    ivals, idevs = list(
        zip(*[tuple_fromstr(c) for c in dpts["intensities"].pop("#text").split()])
    )
    iunit = dpts["intensities"].pop("@unit")
    timestamp = header.pop("startTimeStamp")

    dp = {
        "intensity": {"vals": ivals, "devs": idevs, "unit": iunit},
        "timestamp": timestamp,
        "counting_time": counting_time,
    }

    positions = _process_values(dpts.pop("positions"))
    for v in positions:
        pos = np.linspace(
            float(v["startPosition"]), float(v["endPosition"]), num=len(ivals)
        )
        adiff = np.abs(np.diff(pos)) * 0.5
        adiff = np.append(adiff, adiff[-1])
        dp[v["@axis"]] = {
            "vals": pos,
            "devs": adiff,
            "unit": v["@unit"],
        }
    return dp


def _process_comment(comment: dict) -> dict:
    entry = comment.pop("entry")
    ret = {}
    for line in entry:
        ret.update(panalytical_comment(line))
    return ret


def _process_measurement(measurement: dict, timezone: str):
    """
    A function that processes each section of the XRD XML file.
    """
    # Comment.
    comment = measurement["comment"].pop("entry")
    for line in comment:
        if "PHD Lower Level" in line and "PHD Upper Level" in line:
            __, values = list(zip(*[s.split(" = ") for s in line.split(", ")]))
    keys = ["phd_lower_level", "phd_upper_level"]
    measurement["comment"] = dict(zip(keys, values))
    # Wavelength.
    wavelength = _process_values(measurement.pop("usedWavelength"))
    measurement["wavelength"] = wavelength
    # Incident beam path.
    incident_beam_path = _process_values(measurement.pop("incidentBeamPath"))
    measurement["incident_beam_path"] = incident_beam_path
    # Diffracted beam path.
    diffracted_beam_path = _process_values(measurement.pop("diffractedBeamPath"))
    measurement["diffracted_beam_path"] = diffracted_beam_path
    scan = _process_scan(measurement.pop("scan"))
    trace = {"angle": scan.pop("2Theta"), "intensity": scan.pop("intensity")}
    meta = measurement
    meta["counting_time"] = scan.pop("counting_time")
    trace["uts"] = dateutils.str_to_uts(
        timestamp=scan.pop("timestamp"), timezone=timezone
    )
    return trace, meta


@extract.register(Path)
def extract_from_path(
    source: Path,
    *,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    it = ElementTree.iterparse(source)
    # Removing xmlns prefixes from all tags.
    # From https://stackoverflow.com/a/25920989.
    for __, e in it:
        __, xmlns_present, postfix = e.tag.partition("}")
        if xmlns_present:
            e.tag = postfix  # Strip away all xmlns prefixes.
    root = it.root
    xrd = etree_to_dict(root)
    # Start processing the xml contents.
    measurements = xrd["xrdMeasurements"]
    assert measurements["@status"] == "Completed", "Incomplete measurement."
    comment = _process_comment(measurements["comment"])
    # Renaming some entries because I want to.
    sample = measurements["sample"]
    sample["prepared_by"] = sample.pop("preparedBy")
    sample["type"] = sample.pop("@type")
    # Process measurement data.
    data, meta = _process_measurement(measurements["xrdMeasurement"], timezone)
    data["fn"] = str(source)
    # Shove unused data into meta
    meta["sample"] = sample
    meta["comment"] = comment
    meta["fulldate"] = True
    # Build Datasets
    vals = xr.Dataset(
        data_vars={
            "intensity": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["vals"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "ancillary_variables": "intensity_std_err",
                },
            ),
            "intensity_std_err": (
                ["uts", "angle"],
                np.reshape(data["intensity"]["devs"], (1, -1)),
                {
                    "units": data["intensity"]["unit"],
                    "standard_name": "intensity standard_error",
                },
            ),
            "angle_std_err": (
                ["uts", "angle"],
                np.reshape(data["angle"]["devs"], (1, -1)),
                {
                    "units": data["angle"]["unit"],
                    "standard_name": "angle standard_error",
                },
            ),
        },
        coords={
            "uts": (["uts"], [data["uts"]]),
            "angle": (
                ["angle"],
                data["angle"]["vals"],
                {
                    "units": data["angle"]["unit"],
                    "ancillary_variables": "angle_std_err",
                },
            ),
        },
        attrs=dict(original_metadata=meta),
    )
    return DataTree(vals)



"""
For processing Inficon Fusion json data.  Contains both the data from the raw
chromatogram and the post-processed results.

Usage
`````
Available since ``yadg-4.0``.

.. autopydantic_model:: dgbowl_schemas.yadg.dataschema_6_0.filetype.Fusion_json

Schema
``````
.. code-block:: yaml

    xarray.DataTree:
      coords:
        uts:              !!float
        species:          !!str
      data_vars:
        height:           (uts, species)        # Peak height at maximum
        area:             (uts, species)        # Integrated peak area
        concentration:    (uts, species)        # Calibrated peak area
        xout:             (uts, species)        # Mole fraction (normalized conc.)
        retention time:   (uts, species)        # Peak retention time
      {{ detector_name }}:
        coords:
          uts:            !!float               # Unix timestamp
          elution_time:   !!float               # Elution time
        data_vars:
          signal:         (uts, elution_time)   # Signal data
          valve:          (uts)                 # Valve position

Metadata
````````
No metadata is currently extracted.

.. codeauthor::
    Peter Kraus

"""

import json
import logging
from xarray import Dataset, DataTree
import xarray as xr
import numpy as np
from yadg import dgutils
from pathlib import Path
from yadg.extractors import get_extract_dispatch

extract = get_extract_dispatch()


logger = logging.getLogger(__name__)


def chromdata(jsdata: dict, uts: float) -> Dataset:
    metadata = {}
    metadata["method"] = jsdata["methodName"]
    metadata["version"] = jsdata["softwareVersion"]["version"]
    if jsdata.get("sequence") is not None:
        metadata["datafile"] = jsdata["sequence"].get("location")
    if jsdata.get("annotations") is not None:
        metadata["sampleid"] = jsdata["annotations"].get("name", None)

    units = {
        "height": None,
        "area": None,
        "concentration": "%",
        "xout": "%",
        "retention time": "s",
    }

    raw = {
        "height": {},
        "area": {},
        "concentration": {},
        "xout": {},
        "retention time": {},
    }

    species = set()

    # sort detector keys to ensure alphabetic order for ID matching
    for detname in sorted(jsdata["detectors"].keys()):
        detdict = jsdata["detectors"][detname]
        if "analysis" in detdict:
            for peak in detdict["analysis"]["peaks"]:
                if "label" not in peak:
                    continue
                else:
                    species.add(peak["label"])
                if "height" in peak:
                    if peak["height"] is None:
                        continue
                    raw["height"][peak["label"]] = (float(peak["height"]), 1.0)
                if "area" in peak:
                    if peak["area"] is None:
                        continue
                    raw["area"][peak["label"]] = (float(peak["area"]), 0.01)
                if "concentration" in peak:
                    if peak["concentration"] is None:
                        continue
                    raw["concentration"][peak["label"]] = (
                        float(peak["concentration"]),
                        float(peak["concentration"]) * 1e-3,
                    )
                if "normalizedConcentration" in peak:
                    if peak["normalizedConcentration"] is None:
                        continue
                    raw["xout"][peak["label"]] = (
                        float(peak["normalizedConcentration"]),
                        float(peak["normalizedConcentration"]) * 1e-3,
                    )
                if "top" in peak:
                    if peak["top"] is None:
                        continue
                    raw["retention time"][peak["label"]] = (float(peak["top"]), 0.01)

    valve = jsdata.get("annotations", {}).get("valcoPosition", None)
    if valve is not None:
        raw["valve"] = valve

    species = sorted(species)
    data_vars = {}
    if len(species) > 0:
        for k, v in units.items():
            vals, devs = zip(*[raw[k].get(s, (np.nan, np.nan)) for s in species])
            data_vars[k] = (
                ["uts", "species"],
                [vals],
                {"ancillary_variables": f"{k}_std_err"},
            )
            data_vars[f"{k}_std_err"] = (
                ["uts", "species"],
                [devs],
                {"standard_name": f"{k} stdandard_error"},
            )
            if v is not None:
                data_vars[k][2]["units"] = v
                data_vars[f"{k}_std_err"][2]["units"] = v

    ds = xr.Dataset(
        data_vars=data_vars,
        coords={"species": (["species"], species), "uts": (["uts"], [uts])},
        attrs=dict(original_metadata=metadata),
    )
    return ds


def chromtrace(jsdata: dict, uts: float) -> DataTree:
    # sort detector keys to ensure alphabetic order for ID matching
    traces = sorted(jsdata["detectors"].keys())
    vals = {}
    for detname in traces:
        detdict = jsdata["detectors"][detname]
        fvals = xr.Dataset(
            data_vars={
                "signal": (
                    ["uts", "elution_time"],
                    [detdict["values"]],
                    {"ancillary_variables": "signal_std_err"},
                ),
                "signal_std_err": (
                    ["uts", "elution_time"],
                    [np.ones(detdict["nValuesExpected"])],
                    {"standard_name": "signal standard_error"},
                ),
                "elution_time_std_err": (
                    ["elution_time"],
                    np.ones(detdict["nValuesExpected"]) / detdict["nValuesPerSecond"],
                    {"units": "s", "standard_name": "elution_time standard_error"},
                ),
            },
            coords={
                "elution_time": (
                    ["elution_time"],
                    np.arange(detdict["nValuesExpected"]) / detdict["nValuesPerSecond"],
                    {"units": "s", "ancillary_variables": "elution_time_std_err"},
                ),
                "uts": (["uts"], [uts]),
            },
            attrs={},
        )
        valve = jsdata.get("annotations", {}).get("valcoPosition", None)
        if valve is not None:
            fvals["valve"] = valve
        vals[detname] = fvals

    dt = DataTree.from_dict(vals)
    return dt


@extract.register(Path)
def extract_from_path(
    source: Path,
    *,
    encoding: str,
    timezone: str,
    **kwargs: dict,
) -> DataTree:
    with open(source, "r", encoding=encoding, errors="ignore") as infile:
        jsdata = json.load(infile)
    uts = dgutils.str_to_uts(timestamp=jsdata["runTimeStamp"], timezone=timezone)
    data = chromdata(jsdata, uts)
    trace = chromtrace(jsdata, uts)
    newdt = DataTree(data)
    for k, v in trace.items():
        newdt[k] = v
    return newdt
